//+------------------------------------------------------------------+
//|                                                   Optimystic.mqh |
//|                                 Copyright  2011-2016, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//| MQL's OOP notes: On The Fly Self-Optimization of Expert Advisers |
//|                        https://www.mql5.com/en/blogs/post/683357 |
//+------------------------------------------------------------------+

enum PERFORMANCE_ESTIMATOR
{
  ESTIMATOR_DEFAULT = -1,
  ESTIMATOR_NET_PROFIT, // default
  ESTIMATOR_PROFIT_FACTOR,
  ESTIMATOR_SHARPE,
  ESTIMATOR_DRAWDOWN_PERCENT,
  ESTIMATOR_WIN_PERCENT,
  ESTIMATOR_AVERAGE_PROFIT,
  ESTIMATOR_TRADES_COUNT,
  ESTIMATOR_CUSTOM
};

enum TIME_UNIT
{
  UNIT_BAR,
  UNIT_DAY
};

class OptimysticParameter
{
  public:
    virtual string getName() const = 0;
    virtual double getMin() const = 0;
    virtual double getMax() const = 0;
    virtual double getStep() const = 0;
    virtual double getValue() const = 0;
    virtual double getBest() const = 0;
    virtual bool getEnabled() const = 0;
    virtual void setEnabled(const bool e) = 0;
};

class OptimysticTradeContext
{
  public:
    virtual datetime TimeCurrent() = 0;
    
    virtual double AccountBalance() = 0;
    virtual double AccountEquity() = 0;
    virtual double AccountProfit() = 0;
    virtual double AccountFreeMargin() = 0;
    virtual double AccountFreeMarginCheck(string symbol, int cmd, double volume) = 0;
    virtual double AccountMargin() = 0;

    virtual int OrderSend(string symbol, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit, string comment = NULL, int magic = 0, datetime expiration = 0, color arrow = clrNONE) = 0;
    virtual bool OrderClose(int ticket, double lots, double price, int slippage, color arrow = clrNONE) = 0;
    virtual bool OrderDelete(int ticket, color arrow = clrNONE) = 0;
    virtual bool OrderModify(int ticket, double price, double stoploss, double takeprofit, datetime expiration, color arrow = clrNONE) = 0;
    virtual bool OrderSelect(int index, int select, int pool = MODE_TRADES) = 0;

    virtual int OrdersHistoryTotal() = 0;
    virtual int OrdersTotal() = 0;
    
    virtual double OrderClosePrice() = 0;
    virtual datetime OrderCloseTime() = 0;
    virtual string OrderComment() = 0;
    virtual double OrderCommission() = 0;
    virtual datetime OrderExpiration() = 0;
    virtual double OrderLots() = 0;
    virtual int OrderMagicNumber() = 0;
    virtual double OrderOpenPrice() = 0;
    virtual datetime OrderOpenTime() = 0;
    virtual void OrderPrint() = 0;
    virtual double OrderProfit() = 0;
    virtual double OrderStopLoss() = 0;
    virtual double OrderSwap() = 0;
    virtual string OrderSymbol() = 0;
    virtual double OrderTakeProfit() = 0;
    virtual int OrderTicket() = 0;
    virtual int OrderType() = 0;
    
    virtual double iOpen(string symbol, int tf, int bar) = 0;
    virtual double iHigh(string symbol, int tf, int bar) = 0;
    virtual double iLow(string symbol, int tf, int bar) = 0;
    virtual double iClose(string symbol, int tf, int bar) = 0;
    virtual datetime iTime(string symbol, int tf, int bar) = 0;
    virtual int iBarShift(string symbol, int tf, datetime dt, bool exact = false) = 0;

    // TODO: iVolume iHighest iLowest iBars
    // TODO: iCustom and all standard indicators
    // TODO: MarketInfo
};

class OptimysticCallback
{
  private:
    OptimysticTradeContext *context;
    
  protected:
    datetime TimeCurrent()
    {
      return context != NULL ? context.TimeCurrent() : ::TimeCurrent();
    }
    
    double AccountBalance()
    {
      return context != NULL ? context.AccountBalance() : ::AccountBalance();
    }
    
    double AccountEquity()
    {
      return context != NULL ? context.AccountEquity() : ::AccountEquity();
    }
    
    double AccountProfit()
    {
      return context != NULL ? context.AccountProfit() : ::AccountProfit();
    }
    
    double AccountFreeMargin()
    {
      return context != NULL ? context.AccountFreeMargin() : ::AccountFreeMargin();
    }
    
    double AccountFreeMarginCheck(string symbol, int cmd, double volume)
    {
      return context != NULL ? context.AccountFreeMarginCheck(symbol, cmd, volume) : ::AccountFreeMarginCheck(symbol, cmd, volume);
    }
    
    double AccountMargin()
    {
      return context != NULL ? context.AccountMargin() : ::AccountMargin();
    }
    
    
    int OrdersHistoryTotal()
    {
      return context != NULL ? context.OrdersHistoryTotal() : ::OrdersHistoryTotal();
    }
    
    int OrdersTotal()
    {
      return context != NULL ? context.OrdersTotal() : ::OrdersTotal();
    }
    
    int OrderSend(string symbol, int cmd, double volume, double price, int deviation, double stoploss, double takeprofit, string comment = NULL, int magic = 0, datetime expiration = 0, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderSend(symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magic, expiration, arrow) : ::OrderSend(symbol, cmd, volume, price, deviation, stoploss, takeprofit, comment, magic, expiration, arrow);
    }
    
    bool OrderClose(int ticket, double lots, double price, int slippage, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderClose(ticket, lots, price, slippage, arrow) : ::OrderClose(ticket, lots, price, slippage, arrow);
    }
    
    bool OrderDelete(int ticket, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderDelete(ticket, arrow) : ::OrderDelete(ticket, arrow);
    }
    
    bool OrderModify(int ticket, double price, double stoploss, double takeprofit, datetime expiration, color arrow = clrNONE)
    {
      return context != NULL ? context.OrderModify(ticket, price, stoploss, takeprofit, expiration, arrow) : ::OrderModify(ticket, price, stoploss, takeprofit, expiration, arrow);
    }
    
    bool OrderSelect(int index, int select, int pool = MODE_TRADES)
    {
      return context != NULL ? context.OrderSelect(index, select, pool) : ::OrderSelect(index, select, pool);
    }

    double OrderClosePrice()
    {
      return context != NULL ? context.OrderClosePrice() : ::OrderClosePrice();
    }
    
    datetime OrderCloseTime()
    {
      return context != NULL ? context.OrderCloseTime() : ::OrderCloseTime();
    }

    string OrderComment()
    {
      return context != NULL ? context.OrderComment() : ::OrderComment();
    }
    
    double OrderCommission()
    {
      return context != NULL ? context.OrderCommission() : ::OrderCommission();
    }
    
    datetime OrderExpiration()
    {
      return context != NULL ? context.OrderExpiration() : ::OrderExpiration();
    }
    
    double OrderLots()
    {
      return context != NULL ? context.OrderLots() : ::OrderLots();
    }
    
    int OrderMagicNumber()
    {
      return context != NULL ? context.OrderMagicNumber() : ::OrderMagicNumber();
    }
    
    double OrderOpenPrice()
    {
      return context != NULL ? context.OrderOpenPrice() : ::OrderOpenPrice();
    }
    
    datetime OrderOpenTime()
    {
      return context != NULL ? context.OrderOpenTime() : ::OrderOpenTime();
    }
    
    void OrderPrint()
    {
      if(context != NULL) context.OrderPrint();
      else ::OrderPrint();
    }
    
    double OrderProfit()
    {
      return context != NULL ? context.OrderProfit() : ::OrderProfit();
    }
    
    double OrderStopLoss()
    {
      return context != NULL ? context.OrderStopLoss() : ::OrderStopLoss();
    }
    
    double OrderSwap()
    {
      return context != NULL ? context.OrderSwap() : ::OrderSwap();
    }
    
    string OrderSymbol()
    {
      return context != NULL ? context.OrderSymbol() : ::OrderSymbol();
    }
    
    double OrderTakeProfit()
    {
      return context != NULL ? context.OrderTakeProfit() : ::OrderTakeProfit();
    }
    
    int OrderTicket()
    {
      return context != NULL ? context.OrderTicket() : ::OrderTicket();
    }
    
    int OrderType()
    {
      return context != NULL ? context.OrderType() : ::OrderType();
    }

    double iOpen(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iOpen(symbol, tf, bar) : ::iOpen(symbol, tf, bar);
    }

    double iHigh(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iHigh(symbol, tf, bar) : ::iHigh(symbol, tf, bar);
    }

    double iLow(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iLow(symbol, tf, bar) : ::iLow(symbol, tf, bar);
    }

    double iClose(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iClose(symbol, tf, bar) : ::iClose(symbol, tf, bar);
    }
    
    datetime iTime(string symbol, int tf, int bar)
    {
      return context != NULL ? context.iTime(symbol, tf, bar) : ::iTime(symbol, tf, bar);
    }
    
    int iBarShift(string symbol, int tf, datetime dt, bool exact = false)
    {
      return context != NULL ? context.iBarShift(symbol, tf, dt, exact) : ::iBarShift(symbol, tf, dt, exact);
    }
    

  public:
    void setContext(OptimysticTradeContext *c)
    {
      context = c;
    }
    
    OptimysticTradeContext *getContext() const
    {
      return context;
    }
    
    virtual bool onBeforeOptimization() = 0;
    virtual bool onApplyParameters() = 0;
    virtual void trade(const int bar, const double Ask, const double Bid) = 0;
    virtual void onReadyParameters()
    {
    }
    virtual double onCustomEstimate()
    {
      return EMPTY_VALUE;
    }
    virtual void onAfterOptimization() = 0;
};

class Optimystic
{
  public:
    virtual OptimysticParameter *addParameter(const string name, const double min, const double max, const double step) = 0;
    virtual OptimysticParameter *getParameter(const int i) const = 0;
    OptimysticParameter *operator[](int i) const
    {
      return getParameter(i);
    }
    virtual void setHistoryDepth(const int size, const TIME_UNIT unit) = 0;
    virtual void setReoptimizationPeriod(const int size, const TIME_UNIT unit) = 0;
    virtual void setOptimizationCriterion(const PERFORMANCE_ESTIMATOR type) = 0;
    virtual void setDeposit(const double deposit) = 0;
    virtual void setSpread(const int points = 0) = 0;
    virtual void setCommission(const double c) = 0;
    virtual void setEnabled(const bool enabled) = 0;
  
    virtual double getOptimalPerformance(PERFORMANCE_ESTIMATOR type = ESTIMATOR_DEFAULT) = 0;
    
    virtual void onTick() = 0;
};


#ifndef OPTIMYSTIC_LIBRARY
#import "Optimystic.ex4"
  Optimystic *createInstance(OptimysticCallback &callback);
#import
#endif
